/**
 * pyramid -- implements a simple solitaire game.
 * --
 * $Id: Pyramid.java,v 1.47 1997/05/28 17:29:57 teto Exp teto $
 *
 * This file is part of "The Java Pyramid Solitaire Applet".
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * "The Java Pyramid Solitaire Applet" is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").
 * ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES
 * IS SPECIFICALLY DISCLAIMED.
 */

//
// Feedback about this code is welcome - send email to t_e_t_o@geocities.com
//

// TODO:
// redo backGround color scheme
//
// fix jumping button
//
// fix history when only one card exists in crib
//
// add intro / help screen
// To play: ace = 1, king = 13, match any two EXPOSED cards (on
// pyramid bottom or on crib) to remove them.
//
// add new 50% cardset - remove scaling (also 75% ?)
//
// add prize for winning
//

import java.applet.Applet;
import java.awt.*;
import java.awt.image.*;
import java.util.Random;
import java.util.Vector;

class Move implements Constants {
    int c1, c2, r1, r2;

    Move(int C1, int C2, int R1, int R2) {
	c1 = C1;
	c2 = C2;
	r1 = R1;
	r2 = R2;
    }

    public boolean isEqual(Move m) {
	if (c1 == m.c1 && c2 == m.c2 && r1 == m.r1 && r2 == m.r2)
	    return true;
	else
	    return false;
    }

    public boolean isEqual(int C1, int C2, int R1, int R2) {
	if (c1 == C1 && c2 == C2 && r1 == R1 && r2 == R2)
	    return true;
	else
	    return false;
    }
}

class History implements Constants {
    private int stackPtr, lastCard;
    private Vector moves;
    private Crib crib;
    private Shuffle deck;
    private boolean DEBUG;

    History(Crib cr, Shuffle dk, boolean db) {
	crib = cr;
	DEBUG = db;
	moves = new Vector();
	restart(dk);
    }

    public void restart(Shuffle dk) {
	deck = dk;
	if (moves != null)
	    moves.setSize(0);

	stackPtr = 0;
	newMove(NO_MOVE, NO_MOVE);
    }

    public void newMove(int r1, int r2) {
	setLastCard(r1);
	newMove(r2);
    }

    public void newMove(int r2) {
	int c1 = crib.getCrib(1), c2 = crib.getCrib(2);
	if (DEBUG)
	    System.out.println("stackPtr: " + stackPtr + " c1: " + c1 + " c2: " + c2 + " r1: " + lastCard + " r2: " + r2);

	if (moves == null || stackPtr == moves.size()) {
	    //
	    //	stackPtr is at end of Vector, thus
	    //	add new element at end of Vector
	    //
	    moves.addElement(new Move(c1, c2, lastCard, r2));
	} else
	    if (!(getElement(stackPtr).isEqual(c1, c2, lastCard, r2))) {
		//
		//	A new branch is started.
		//	stackPtr is before the end of Vector,
		//	a new move is added, but it is different
		//	from the previously recorded move.  Thus,
		//	substitute this move for the old one and
		//	discard any successive moves.
		//
		moves.setElementAt(new Move(c1, c2, lastCard, r2), stackPtr);
		moves.setSize(stackPtr + 1);
	    }

	stackPtr++;
    }

    public int getLastCard() {
	return(lastCard);
    }

    public void setLastCard(int value) {
	lastCard = value;
    }

    public int getPtr() {
	return(stackPtr);
    }

    public int getSize() {
	return(moves.size());
    }

    private Move getElement(int element) {
	if (element >= 0 && element < moves.size())
	    return ((Move) moves.elementAt(element));
	else
	    return null;
    }

    public int back() {
	Move l, p;

	if (stackPtr > 1) {
	    if (DEBUG)
		System.out.println("before BACK stackPtr: " + stackPtr + " c1: " + crib.getCrib(1) + " c2: " + crib.getCrib(2) + " lastCard: " + lastCard);
	    stackPtr--;
	    //
	    //	l = last displayed move.
	    //	p = move previous to last displayed move.
	    //
	    l = (Move) moves.elementAt(stackPtr);
	    p = (Move) moves.elementAt(stackPtr - 1);

	    crib.setCrib(p.c1, 1);
	    crib.setCrib(p.c2, 2);

	    if (DEBUG)
		System.out.println("l.r2: " + l.r2 + " l.r1: " + l.r1 + " p.r1: " + p.r1);
	    if (l.r2 != NO_MOVE) {			// card(s) removed
		deck.setPendClr(l.r2);
		deck.unSetRemoved(l.r2);

		if (l.r1 != l.r2) {			// not a king
		    deck.setSelected(l.r1);
		    lastCard = l.r1;
		}
	    } else {					// not a remove (l.r2)
		if (l.r1 != p.r1) {			// something changed
		    if (l.r1 != NO_MOVE) {		// not an unselect
			deck.setPendClr(l.r1);
			deck.unSetSelected(l.r1);
			lastCard = NO_MOVE;
		    } else {				// an unselect
			if (p.r1 != p.r2) {		// not a king
			    deck.setSelected(p.r1);
			    lastCard = p.r1;
			}
		    }
		}
	    }
	    if (DEBUG)
		System.out.println("after BACK stackPtr: " + stackPtr + " c1: " + crib.getCrib(1) + " c2: " + crib.getCrib(2) + " lastCard: " + lastCard);
	}

	return(stackPtr);
    }

    public int forward() {
	Move l, n;

	if (stackPtr < moves.size()) {
	    if (DEBUG)
		System.out.println("before FORWARD stackPtr: " + stackPtr + " c1: " + crib.getCrib(1) + " c2: " + crib.getCrib(2) + " lastCard: " + lastCard);
	    stackPtr++;
	    //
	    //	l = last displayed move.
	    //	n = next move after last displayed move.
	    //
	    l = (Move) moves.elementAt(stackPtr - 2);
	    n = (Move) moves.elementAt(stackPtr - 1);

	    crib.setCrib(n.c1, 1);
	    crib.setCrib(n.c2, 2);

	    if (n.r2 != NO_MOVE) {			// card(s) removed
		deck.setRemoved(n.r2);

		if (n.r1 != n.r2) {			// not a king
		    deck.setRemoved(n.r1);
		    lastCard = NO_MOVE;
		}
	    } else {					// not a remove (n.r2)
		if (n.r1 != l.r1) {			// something changed
		    if (n.r1 != NO_MOVE) {		// not an unselect
			deck.setSelected(n.r1);
			lastCard = n.r1;
		    } else {				// an unselect
			if (l.r1 != l.r2) {		// last move not a king
			    deck.setPendClr(l.r1);
			    deck.unSetSelected(l.r1);
			    lastCard = NO_MOVE;
			}
		    }
		}
	    }
	    if (DEBUG)
		System.out.println("after FORWARD stackPtr: " + stackPtr + " c1: " + crib.getCrib(1) + " c2: " + crib.getCrib(2) + " lastCard: " + lastCard);
	}
	return(stackPtr);
    }
}

class HistoryCanvas extends Panel implements Constants {
    private History history;
    private boolean DEBUG;
    private Pyramid o;
    private Dimension d;
    private int radius, status[];
    private Point ctrPts[][];

    private final int BACKWARD = 0;
    private final int FORWARD = 1;

    HistoryCanvas(History h, boolean debug, Pyramid theOwner, Dimension d) {
	history = h;
	DEBUG = debug;
	o = theOwner;
	this.d = d;
	radius = d.width / 6;
	status = new int[2];
	ctrPts = new Point[2][4];
	//
	for (int i = 0, sign = 1; i < 2; i++) {
	    ctrPts[i][0] = new Point((d.width / 2) - sign * (int)(radius * 1.25), d.height / 4);
	    ctrPts[i][1] = new Point(ctrPts[i][0].x   - sign * radius, ctrPts[i][0].y);
	    ctrPts[i][2] = new Point(ctrPts[i][0].x   + sign * (radius / 2), ctrPts[i][0].y + (int) (radius * 0.866));
	    ctrPts[i][3] = new Point(ctrPts[i][0].x   + sign * (radius / 2), ctrPts[i][0].y - (int) (radius * 0.866));
	    sign = -1;
	}
	//
	restart();
    }

    public void restart() {
	setBackground(Color.pink);
	status[BACKWARD] = 0;
	status[FORWARD] = 0;
	update();
    }

    private int whichIcon(int x, int y) {
	for (int i = 0; i < 2; i++) {
	    int l1 = Math.abs(ctrPts[i][0].x - x);
	    int l2 = Math.abs(ctrPts[i][0].y - y);
	    int nr = (int) Math.sqrt(l1 * l1 + l2 * l2);
	    if (nr <= radius)
		return i;
	}
	return NULL;
    }

    public void update() {
        int size = history.getSize() - 1;
        int ptr = history.getPtr() - 1;
        boolean repaint = false;

	if ((ptr > 0 && status[0] == 0) || (ptr == 0 && status[0] > 0)) {
	    status[0] = ptr;
	    repaint = true;
	}

	if ((size > ptr && status[1] == 0) || (size == ptr && status[1] > 0)) {
	    status[1] = size - ptr;
	    repaint = true;
	}

	if (DEBUG)
	    System.out.println("history current: " + ptr + " max: " + size);

	if (repaint)
	    repaint();
    }

    public boolean handleEvent(Event evt) {
        int size = history.getSize() - 1;
        int ptr = history.getPtr() - 1;
        boolean update = false;

        switch (evt.id) {
	    case Event.MOUSE_DOWN:
                switch(whichIcon(evt.x, evt.y)) {
		    case FORWARD:
			if (ptr < size) {
			    history.forward();
			    update = true;
			}
			break;
		    case BACKWARD:
			if (ptr > 0) {
			    history.back();
			    update = true;
			}
			break;
		    default:
			break;
		}

		if (DEBUG)
		    System.out.println("maxHist: " + size + " getPtr: " + history.getPtr());

		o.update(true, true, update);
                break;
	    default:
                break;
	}
	return true;
    }

    public Dimension minimumSize() {
	return d;
    }

    public Dimension preferredSize() {
	return minimumSize();
    }

    public void update(Graphics g) {
	paint(g);
    }

    public void paint(Graphics g) {
	for (int i = 0; i < 2; i++) {
	    if (status[i] > 0)
		g.setColor(Color.green);
	    else
		g.setColor(Color.red);
	    //
	    g.fillOval(ctrPts[i][0].x - radius, ctrPts[i][0].y - radius,
		radius * 2, radius * 2);
	    //
	    int exes[] = { ctrPts[i][1].x, ctrPts[i][2].x, ctrPts[i][3].x };
	    int eyes[] = { ctrPts[i][1].y, ctrPts[i][2].y, ctrPts[i][3].y };
	    g.setColor(Color.black);
	    g.fillPolygon(exes, eyes, 3);
	}
    }
}

class MorphColor implements Constants {
    private final int DELTA = 5;
    private Color nextColor;
    private int rDelta, gDelta, bDelta;
    private boolean DEBUG;

    MorphColor(Color oldColor, boolean db) {
	DEBUG = db;
	colorRoll();
	colorChange(oldColor);
    }

    MorphColor(boolean db) {
	DEBUG = db;
	colorRoll();
	nextColor = colorRandom();
    }

    /**
     * generate a random color
     */
    public Color colorRandom() {
	nextColor = new Color(
	    (int)Math.floor(Math.random() * 256),
	    (int)Math.floor(Math.random() * 256),
	    (int)Math.floor(Math.random() * 256)
	);
	return(getColor());
    }

    /**
     * reselect color delta's
     */
    private void colorRoll() {
	rDelta = (int)Math.floor(Math.random() * DELTA) - (DELTA % 2);
	gDelta = (int)Math.floor(Math.random() * DELTA) - (DELTA % 2);
	bDelta = (int)Math.floor(Math.random() * DELTA) - (DELTA % 2);
	if (rDelta == 0 && gDelta == 0 && bDelta == 0)
	    colorRoll();
    }

    /**
     * change/progress/fade the color
     */
    private void colorChange(Color color) {
	int rColor = color.getRed();
	int gColor = color.getGreen();
	int bColor = color.getBlue();

	rColor += DELTA * rDelta;
	if ( rColor > 255 || rColor < 0) {
	    rColor -= DELTA * rDelta;
	    colorRoll();
	}
	gColor += DELTA * gDelta;
	if ( gColor > 255 || gColor < 0) {
	    gColor -= DELTA * gDelta;
	    colorRoll();
	}
	bColor += DELTA * bDelta;
	if ( bColor > 255 || bColor < 0) {
	    bColor -= DELTA * bDelta;
	    colorRoll();
	}
	nextColor = new Color(rColor, gColor, bColor);
    }

    public Color getColor() {
	if (DEBUG)
	    System.out.println("Color red: " + nextColor.getRed() + " green: " + nextColor.getGreen() + " blue: " + nextColor.getBlue());
	return(nextColor);
    }

    public Color getNextColor() {
	colorChange(nextColor);
	return(getColor());
    }
}

class NextPoint extends Panel implements Constants {
    private final int DELTA = 10;
    private int dirx = 1, diry = 1;
    private Dimension d;
    private Dimension c;
    private Point ctr;

    NextPoint(int imgx, int imgy) {
	restart(imgx, imgy);
    }

    public void restart(int imgx, int imgy) {
	d = new Dimension(size());
	setCSize(imgx, imgy);
	setCenter();
    }

    private void setCSize(int imgx, int imgy) {
	c = new Dimension(imgx, imgy);
    }

    private void setCenter() {
	ctr = new Point((d.width - c.width) / 2, (d.height - c.height) / 2);
    }

    public Point getCenter() {
	return(ctr);
    }

    Point randomPoint() {
	Point p = new Point(
	    (int)Math.floor(Math.random() * (d.width  - c.width)),
	    (int)Math.floor(Math.random() * (d.height - c.height))
	);
	return(p);
    }

    Point advancePoint(Point orgPoint) {
	int currx = orgPoint.x + (dirx * DELTA);
	int curry = orgPoint.y + (diry * DELTA);

	if ((currx > d.width - c.width) || (currx < 0)) {
	    dirx *= -1;
	    while ((currx > d.width - c.width) || (currx < 0)) {
		currx += (dirx * DELTA);
	    }
	}

	if ((curry > d.height - c.height) || (curry < 0)) {
	    diry *= -1;
	    while ((curry > d.width - c.height) || (curry < 0)) {
		curry += (diry * DELTA);
	    }
	}

	return(new Point(currx, curry));
    }
}

class IdiotButton extends NextPoint {
    private int jumpy;
    private boolean DEBUG;
    private Pyramid o;
    private Dimension d;
    private boolean wild = false;
    private Button restartButton;
    private Font cntFont;

    IdiotButton(int j, boolean debug, Pyramid theOwner, Dimension d, int scale) {
	super(0, 0);
	jumpy = j;
	DEBUG = debug;
	o = theOwner;
	this.d = d;
	cntFont = new Font("Helvetica", Font.PLAIN, (int) ((16 * scale) / 100));
	setFont(cntFont);
	restartButton = new Button("New Game");
	add(restartButton);
	restart();
    }

    public void restart() {
	setBackground(Color.pink);
	setForeground(Color.black);
	super.restart(restartButton.size().height, restartButton.size().width);
	mvCenter();
    }

    private void mvCenter() {
	Point p = getCenter();
	restartButton.move(p.x, p.y);
    }

    public Dimension minimumSize() {
	return d;
    }

    public Dimension preferredSize() {
	return minimumSize();
    }

    public boolean mouseMove(Event event, int i, int j) {
	if (! wild)
	    return true;
	Point p = randomPoint();
	restartButton.move(p.x, p.y);
	return true;
    }

    public boolean mouseExit(Event evt, int x, int y) {
	mvCenter();
	return true;
    }

    public boolean action(Event event, Object object) {
	if (!(event.target instanceof Button))
	    return false;
	if (event.target == restartButton) {
	    if (wild) {
		o.restart();
	    } else
		wild = true;
	}
	return true;
    }
}

class PyramidLayout implements Constants {
    private Point[] cardLayout;
    private int[] row;
    private int[] col;
    private int CELLX, CELLY;
    int length;

    PyramidLayout(int cx, int cy) {
	int currentX, currentY, centerX;

	CELLX = cx;
	CELLY = cy;
	centerX = (((CELLX + PADD) * ROWS) / 2) + PADD;
	length = cellCount(ROWS);

	cardLayout = new Point[length];
	row = new int[length];
	col = new int[length];

	for (int i = 1, ctr = 0; i <= ROWS; i++) {
	    currentX = centerX - ((i * (CELLX + PADD)) / 2);
	    currentY = (i - 1) * (CELLY + PADD);
	    for (int j = 1; j <= i; j++) {
		cardLayout[ctr] = new Point(currentX, currentY);
		row[ctr] = i;
		col[ctr] = j;
		currentX += CELLX + PADD;
		ctr++;
	    }
	}
    }

    int cellCount(int ROWS) {
	int total = 0;

	for (int cnt = 1; cnt <= ROWS; cnt++) {
	    total += cnt;
	}

	return(total);
    }

    int whichCell(int coordX, int coordY) {
	int which = NULL;
	for (int i = 0; i < length; i++) {
	    if ((cardLayout[i].x <= coordX) && (cardLayout[i].x + CELLX >= coordX) &&
		(cardLayout[i].y <= coordY) && (cardLayout[i].y + CELLY >= coordY)) {
		which = i;
		break;
	    }
	}
	return(which);
    }

    Point whichPoint(int cell) {
	if (cell >= length || cell < 0) return(null);
	return(cardLayout[cell]);
    }

    private int whichRow(int cell) {
	if (cell >= length || cell < 0) return(NULL);
	return(row[cell]);
    }

    private int whichCol(int cell) {
	if (cell >= length || cell < 0) return(NULL);
	return(col[cell]);
    }

    int below(int cell, int offset) {
	int nextRow;

	if (cell >= length || cell < 0) return(NULL);

	//
	//	The whichRow(cell) is the last row of the pyramid
	//
	if (cellCount(whichRow(cell)) == length)
	    return(length);

	nextRow = whichRow(cell) + 1;

	for (int i = 0; i < length; i++) {
	    if (nextRow == whichRow(i))
		if ((whichCol(cell) + offset) == whichCol(i))
		    return(i);
	}
	//
	//	This line should not be reached
	//
	return(NULL);
    }

    int belowLeft(int cell) {
	return(below(cell, 0));
    }

    int belowRight(int cell) {
	return(below(cell, 1));
    }

}

class Crib implements Constants {
    Point center0, below2;
    private int CELLX, CELLY, offsetY;
    private Point[] stack = new Point[3];
    private int[] crib = new int[3];

    Crib(int cx, int cy) {
	CELLX = cx;
	CELLY = cy;
	offsetY = 0;
	//
	for (int i = 0; i < 3; i++) {
	    stack[i] = new Point(0, offsetY + (CELLY + PADD) * i);
	}
	//
	center0 = new Point((CELLX / 4), offsetY + (int) (CELLY / 1.5));
	below2  = new Point((CELLX / 3), offsetY + (int) (CELLY * 3.5));
	//
	restart();
    }

    public void restart() {
	setCrib(NULL, 0);
	setCrib(LAYOUT_LENGTH + 1, 1);
	setCrib(LAYOUT_LENGTH, 2);
    }

    int getCrib(int crib) {
	if (crib == 0)
	    return(getCrib0());
	if (crib == 1)
	    return(getCrib1());
	if (crib == 2)
	    return(getCrib2());
	return(NULL);
    }

    private int getCrib0() { return(crib[0]); }
    private int getCrib1() { return(crib[1]); }
    private int getCrib2() { return(crib[2]); }

    int whichCrib(int value) {
	if (isCrib(value, 0))
	    return(0);
	if (isCrib(value, 1))
	    return(1);
	if (isCrib(value, 2))
	    return(2);
	return(NULL);
    }

    private boolean isCrib(int value, int crib) {
	if (crib == 0)
	    return(isCrib0(value));
	if (crib == 1)
	    return(isCrib1(value));
	if (crib == 2)
	    return(isCrib2(value));
	return(false);
    }

    private boolean isCrib0(int value) { return(crib[0] == value); }
    private boolean isCrib1(int value) { return(crib[1] == value); }
    private boolean isCrib2(int value) { return(crib[2] == value); }

    Point getStack(int stack) {
	if (stack == 0)
	    return(getStack0());
	if (stack == 1)
	    return(getStack1());
	if (stack == 2)
	    return(getStack2());
	return(null);
    }

    private Point getStack0() { return(stack[0]); }
    private Point getStack1() { return(stack[1]); }
    private Point getStack2() { return(stack[2]); }

    void setCrib(int num, int crib) {
	if (crib == 0)
	    setCrib0(num);
	else
	    if (crib == 1)
		setCrib1(num);
	    else
		if (crib == 2)
		    setCrib2(num);
    }

    private void setCrib0(int num) { crib[0] = num; }
    private void setCrib1(int num) { crib[1] = num; }
    private void setCrib2(int num) { crib[2] = num; }

    int whichCell(int coordX, int coordY) {
	int which = NULL;
	for (int i = 0; i < 3; i++) {
	    if ((stack[i].x <= coordX) && (stack[i].x + CELLX >= coordX) &&
		(stack[i].y <= coordY) && (stack[i].y + CELLY >= coordY)) {
		which = i;
		break;
	    }
	}
	return(which);
    }
}

class Shuffle implements Constants {
    private int[] shuffled;
    int[] state;

    final int SELECTED	= 1;
    final int REMOVED	= 2;
    final int CLEAR_ST	= 3;
    final int PENDING	= 4;
    final int CLEAR_ALL	= 7;

    Shuffle(int size) {
	shuffled = new int[size];
	state = new int[size];
	int CNT, NDX, PTR, position;
	Random rand = new Random();

	for (NDX = 0; NDX < size; NDX++) {
	    shuffled[NDX] = -1;
	}

	for (NDX = size; NDX > 0; NDX--) {
	    rand.setSeed(rand.nextLong());
	    position = -1;
	    while (position < 0)
		position = (int)Math.floor(rand.nextLong() % NDX);
	    //
	    //	locate the position'th free slot
	    //
	    for (PTR = CNT = 0; CNT <= position; PTR++) {
		if (shuffled[PTR] == -1) {
		    CNT++;
		}
	    }
	    //
	    //	shuffled[] values are to be used as array subscripts...
	    //
	    shuffled[PTR - 1] = NDX - 1;
	}
    }

    int getCard(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	return(shuffled[ndx]);
    }

    int cardValue(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	return((getCard(ndx) + 1) % 13);
    }

    void setSelected(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	state[ndx] = (state[ndx] & PENDING) | SELECTED;
    }

    void setRemoved(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	state[ndx] = (state[ndx] & PENDING) | REMOVED;
    }

    void setPendClr(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	state[ndx] = (state[ndx] & PENDING) | PENDING;
    }

    boolean isSelected(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if ((state[ndx] & SELECTED) == SELECTED) return true;
	else return false;
    }

    boolean isRemoved(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if ((state[ndx] & REMOVED) == REMOVED) return true;
	else return false;
    }

    boolean isPendClr(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if ((state[ndx] & PENDING) == PENDING) return true;
	else return false;
    }

    void unSetSelected(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if (isSelected(ndx))
	    state[ndx] = state[ndx] & (~ SELECTED);
    }

    void unSetRemoved(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if (isRemoved(ndx))
	    state[ndx] = state[ndx] & (~ REMOVED);
    }

    void unSetPendClr(int ndx) {
	// if (ndx < state.length || ndx >= 0)
	if (isPendClr(ndx))
	    state[ndx] = state[ndx] & (~ PENDING);
    }

    int numFree(int first) {
	int i, free = 0;

	// if (first < state.length || first >= 0)
	for (i = first + 1; i < state.length; i++)
	    if (! isRemoved(i))
		free++;
	//
	return(free);
    }

    //
    //	'first' is the first card in the cycle
    //	'start' is where the search starts from
    //
    //	The search will loop from 'start' to state.length
    //	    and back around from 'first' to 'start'.
    //
    int nextFree(int first, int start) {
	int i;
	// if (start < state.length || start >= 0)
	// if (first < state.length || first >= 0)

	if (start == -1)
	    return(NULL);

	for (i = start + 1; i < state.length; i++)
	    if (! isRemoved(i)) {
		// if (DEBUG)
		//    System.out.println("start" + start + " state.length " + state.length + " i " + i);
		return(i);
	    }
	//
	//	loop back around ( if first < start )
	//
	for (i = first; i <= start; i++)
	    if (! isRemoved(i))
		return(i);

	return(NULL);
    }

}

class Rules implements Constants {
    private Crib crib;
    private Shuffle deck;
    private History history;
    private boolean DEBUG;

    Rules(Crib cr, Shuffle dk, History h, boolean db) {
	crib = cr;
	history = h;
	DEBUG = db;
	restart(dk);
    }

    public void restart(Shuffle d) {
	deck = d;
    }

    public boolean selectCard(int whichCard, boolean calledByCrib) {
	int cardValue, lastValue;
	boolean repaintOther = false;
	int lastCard = history.getLastCard();

	//
	//	need some bounds checking on whichCard
	//
	if (whichCard == NO_MOVE) {
	    if (DEBUG)
		System.out.println("bad click.");
	    return repaintOther;
	}
	if (deck.isRemoved(whichCard))
	    return repaintOther;
	if (crib.whichCrib(whichCard) != NULL && calledByCrib == false)
	    repaintOther = true;
	else
	    if (crib.whichCrib(whichCard) == NULL && calledByCrib == true)
		repaintOther = true;
	cardValue = deck.cardValue(whichCard);
	if (DEBUG)
	    System.out.println("	    cardValue " + cardValue);
	if (deck.isSelected(whichCard)) {
	    //
	    //	Selecting a selected card unselects it
	    //
	    if (DEBUG)
		System.out.println("		the card was selected, unselecting ");
	    deck.setPendClr(whichCard);
	    deck.unSetSelected(whichCard);
	    history.newMove(NO_MOVE, NO_MOVE);
	} else {
	    //
	    //	was a previous card selected ?
	    //
	    if (DEBUG)
		System.out.println("		lastCard: " + lastCard);
	    if (lastCard != NO_MOVE) {
		//
		//	What to update ?
		//
		if (crib.whichCrib(lastCard) != NULL && calledByCrib == false)
		    repaintOther = true;
		else
		    if (crib.whichCrib(lastCard) == NULL && calledByCrib == true)
			repaintOther = true;
		//
		lastValue = deck.cardValue(lastCard);
		if (DEBUG)
		    System.out.println("		lastValue " + lastValue);
		if ((cardValue + lastValue) == 13) {
		    if (DEBUG)
			System.out.println("		mark both as removed" );
		    //
		    //	We no longer need to keep track of these
		    //	two cards - they are no longer in play.
		    //	mark both as removed
		    //
		    deck.setRemoved(lastCard);
		    deck.setRemoved(whichCard);
		    //
		    if (crib.whichCrib(lastCard) != NULL)
			cribWalk(lastCard);
		    if (crib.whichCrib(whichCard) != NULL)
			cribWalk(whichCard);
		    //
		    history.newMove(whichCard);
		    history.setLastCard(NO_MOVE);
		} else {
		    //
		    //	UNmark as selected
		    //
		    if (DEBUG)
			System.out.println("		unmark previous as selected" );
		    deck.setPendClr(lastCard);
		    deck.unSetSelected(lastCard);
		    history.newMove(NO_MOVE, NO_MOVE);
		}
	    } else {
		if (DEBUG)
		    System.out.println("		first card to be selected");
		if (cardValue == 0) {
		    if (DEBUG)
			System.out.println("		    its a king");
		    deck.setRemoved(whichCard);
		    if (crib.whichCrib(whichCard) != NULL)
			cribWalk(whichCard);
		    history.newMove(whichCard, whichCard);
		    history.setLastCard(NO_MOVE);
		} else {
		    if (DEBUG)
			System.out.println("		    its NOT a king");
		    history.setLastCard(whichCard);
		    deck.setSelected(whichCard);
		    history.newMove(whichCard, NO_MOVE);
		}
	    }
	}
	return repaintOther;
    }

    public void cribWalk(int thisCard) {
	switch (crib.whichCrib(thisCard)) {
	    case 2:
		if (DEBUG)
		    System.out.println("		    thisCard was in crib 2" );
		mvCrib1to2();
		break;
	    case 1:
		if (DEBUG)
		    System.out.println("		    thisCard was in crib 1" );
		mvCrib0to1();
		break;
	    case 0:
		if (DEBUG)
		    System.out.println("		    thisCard was in crib 0" );
		mvCrib1to2();
		history.newMove(NO_MOVE);
		break;
	    default:
		if (DEBUG)
		    System.out.println("		    thisCard was not in a crib" );
		break;
	}
    }

    private int mvCrib1to2 () {
	int rval, oldCrib2Card;

	oldCrib2Card = crib.getCrib(2);
	crib.setCrib(crib.getCrib(1), 2);
	//
	if ((rval = mvCrib0to1()) == NULL) {
	    //
	    //	no card was moved into crib1
	    //	move the oldCrib2Card crib1
	    //
	    crib.setCrib(oldCrib2Card, 1);
	} else {
	    //
	    //	Only one card can be selected at a time
	    //	and it has to be visible.  Moving it off
	    //	the crib2 deselects it.
	    //
	    if (deck.isSelected(oldCrib2Card)) {
		deck.unSetSelected(oldCrib2Card);
		history.setLastCard(NO_MOVE);
	    }
	}
	return(rval);
    }

    private int mvCrib0to1 () {
	int currentCard = crib.getCrib(2);
	int nextFree = deck.nextFree(LAYOUT_LENGTH, currentCard);

	if ((nextFree == NULL) || (nextFree == currentCard))
	    return(NULL);

	if (DEBUG)
	    System.out.println("nextFree:  " + nextFree + " currentCard:  " + currentCard);
	//
	//	Useless click on the crib ?
	//
	if (nextFree != NULL) {
	    crib.setCrib(nextFree, 1);
	}

	return(nextFree);
    }

}

class PyramidCanvas extends Canvas implements Constants {
    private Pyramid o;
    private boolean DEBUG;
    private int CARDX, CARDY;
    private PyramidLayout layout;
    private Shuffle deck;
    private Image[] cards;
    private Rules rules;
    private Color Grey = new Color(212, 212, 212), cardColor;

    PyramidCanvas(int cx, int cy, boolean db,
			PyramidLayout LO, Shuffle d,
			Color cc, Pyramid theOwner, Image[] a, Rules r) {
	o = theOwner;
	DEBUG = db;
	CARDX = cx;
	CARDY = cy;
	layout = LO;
	cards = a;
	rules = r;
	restart(d, cc);
    }

    public void restart(Shuffle d, Color cc) {
	deck = d;
	cardColor = cc;

	for (int i = 0; i < LAYOUT_LENGTH; i++)
	    deck.setPendClr(i);

	setBackground(Color.white);
    }

    public Dimension minimumSize() {
	return new Dimension((CARDX + PADD) * 7 + PADD, (CARDY + PADD) * 7);
    }

    public Dimension preferredSize() {
	return minimumSize();
    }

    public void update(Graphics g) {
	paint(g);
    }

    public boolean mouseDown(Event evt, int x, int y) {
	int whichCard = NO_MOVE;
	boolean repaintOther = false;

	if ((whichCard = layout.whichCell(x, y)) != NO_MOVE) {
	    if (DEBUG)
		System.out.println("    whichCard  " + whichCard);
	    //
	    //	Is the card 'exposed' / on the edge ?
	    //
	    if  ( (layout.below(whichCard, 0) == LAYOUT_LENGTH)
		    || ( deck.isRemoved(layout.belowLeft(whichCard))
			    && deck.isRemoved(layout.belowRight(whichCard))
		       )
		)
		repaintOther = rules.selectCard(whichCard, false);
		o.update(true, repaintOther, true);
	}
	return true;
    }

    public void paint(Graphics g) {
	Dimension d;
	int card;
	Point coords;

	if (cards == null)
	    return;

	for (int i = 0; i < LAYOUT_LENGTH; i++) {
	    coords = layout.whichPoint(i);
	    //
	    if (deck.isSelected(i)) {
		g.setColor(Grey);
		g.fillRect(coords.x, coords.y, CARDX, CARDY);
		g.drawImage(cards[deck.getCard(i)],
		    coords.x, coords.y, CARDX, CARDY, this);
		g.setColor(Color.white);
	    } else {
		if (deck.isRemoved(i)) {
		    g.setColor(cardColor);
		    g.fillRect(coords.x, coords.y, CARDX, CARDY);
		} else {
		    if (deck.isPendClr(i)) {
			g.setColor(Color.white);
			g.fillRect(coords.x, coords.y, CARDX, CARDY);
			deck.unSetPendClr(i);
		    }
		    g.drawImage(cards[deck.getCard(i)],
			coords.x, coords.y, CARDX, CARDY, this);
		}
	    }
	}
    }
}

class CribCanvas extends Canvas implements Constants {
    private Pyramid o;
    private boolean DEBUG;
    private int CARDX, CARDY;
    private Crib crib;
    private Shuffle deck;
    private Image[] cards;
    private Rules rules;
    private Color Grey = new Color(212, 212, 212), cardColor;
    private Font cntFont;

    CribCanvas(boolean db, int cx, int cy, int scale, Crib C,
			Pyramid theOwner, Image[] a, Rules r,
			Shuffle d, Color cc) {
	DEBUG = db;
	CARDX = cx;
	CARDY = cy;
	cntFont = new Font("Helvetica", Font.PLAIN, (int) ((36 * scale) / 100));
	crib = C;
	cards = a;
	rules = r;
	o = theOwner;
	restart(d, cc);
    }

    public void restart(Shuffle d, Color cc) {
	deck = d;
	cardColor = cc;
	setBackground(Color.gray);
    }

    public Dimension minimumSize() {
	return new Dimension((CARDX + PADD) * 2, (CARDY + PADD) * 3);
    }

    public Dimension preferredSize() {
	return minimumSize();
    }

    public void update(Graphics g) {
	paint(g);
    }

    public boolean mouseDown(Event evt, int x, int y) {
	int whichCard, whichCrib = NO_MOVE;
	boolean repaintOther = false;

	if ((whichCrib = crib.whichCell(x, y)) != NO_MOVE) {
	    if (DEBUG)
		System.out.println("    whichCrib  " + whichCrib);
	    if (whichCrib == 0) {
		//
		//	the card in crib1 is advanced to crib2
		//
		rules.cribWalk(crib.getCrib(0));
		o.update(repaintOther, true, true);
	    } else {
		whichCard = crib.getCrib(whichCrib);
		if (DEBUG)
		    System.out.println("crib." + whichCrib + " whichCARD: " + crib.getCrib(whichCrib));
		repaintOther = rules.selectCard(whichCard, false);
		o.update(repaintOther, true, true);
	    }
	}
	return true;
    }

    public void paint(Graphics g) {
	Dimension d;
	int card;
	Point coords;

	if (cards == null)
	    return;
	//
	g.setColor(cardColor);
	g.fillRect(crib.getStack(0).x, crib.getStack(0).y, CARDX, CARDY);
	g.setColor(Color.black);
	g.drawRect(crib.getStack(0).x, crib.getStack(0).y, CARDX, CARDY);
	g.setFont(cntFont);
	g.drawString(String.valueOf(deck.numFree(LAYOUT_LENGTH - 1)),
		crib.center0.x, crib.center0.y);
	g.setColor(cardColor);
	for (int i = 1; i <= 2; i++) {
	    coords = crib.getStack(i);
	    card = crib.getCrib(i);
	    if (deck.isSelected(card)) {
		g.setColor(Grey);
		g.fillRect(coords.x, coords.y, CARDX, CARDY);
		g.drawImage(cards[deck.getCard(card)],
		    coords.x, coords.y, CARDX, CARDY, this);
		g.setColor(Color.white);
	    } else {
		if (deck.isRemoved(card)) {
		    g.setColor(cardColor);
		    g.fillRect(coords.x, coords.y, CARDX, CARDY);
		} else {
		    if (deck.isPendClr(card))
			deck.unSetPendClr(card);

		    g.setColor(Color.white);
		    g.fillRect(coords.x, coords.y, CARDX, CARDY);

		    g.drawImage(cards[deck.getCard(card)],
			coords.x, coords.y, CARDX, CARDY, this);
		}
	    }
	}
    }
}

/**
 * Pyramid Solitaire
 *
 */
public class Pyramid extends Applet implements Constants {
    final int CARDS = 52;

    private Shuffle deck;
    private MorphColor newColor;
    private Image[] fiftyTwo = new Image[CARDS];
    private boolean DEBUG;
    private Button restartButton;
    private PyramidLayout layout;
    private PyramidCanvas pcanvas;
    private Crib crib;
    private CribCanvas ccanvas;
    private IdiotButton ibutton;
    private History history;
    private HistoryCanvas hcanvas;
    private Rules rules;
    private int CARDX, CARDY, CELLX;
    private Color nextX;
    private Font cntFont;

    public String getAppletInfo() {
        return "Pyramid Solitaire 1.0\nTrane Consulting, Inc\nMay 1997\nFeedback about this code is welcome.\nSend email to t_e_t_o@geocities.com";
    }

    public void init() {
	String fileName;
	int scale, jumpy;
	Image img;
	MediaTracker tracker;
	CropImageFilter cropFilter;
	GridBagLayout gb = new GridBagLayout();
	GridBagConstraints gc = new GridBagConstraints();

	String at = getParameter("DEBUG");
	DEBUG = (at != null) ? true : false;
	fileName = getParameter("image");
	if (fileName == null) {
	    System.out.println("No card image file specified, aborting");
	    System.exit(1);
	}
	//
	at = getParameter("CELLX");
	CELLX = (at != null) ? Integer.valueOf(at).intValue() : 0;
	if (CELLX == 0 || (CARDS % CELLX) != 0) {
	    System.out.println("parameter 'CELLX' must divide evenly into " + CARDS + ", aborting");
	    System.exit(1);
	}
	//
	at = getParameter("scale");
	scale = (at != null) ? Integer.valueOf(at).intValue() : 100;
	if (scale == 0)
	    scale = 100;
	//
	at = getParameter("jumpy");
	jumpy = (at != null) ? Integer.valueOf(at).intValue() : 100;
	if (jumpy < 1)
	    jumpy = 100;
	//
	cntFont = new Font("Helvetica", Font.PLAIN, (int) ((18 * scale) / 100));
	//
	tracker  =  new  MediaTracker(this);
	//
	showStatus("Getting the card images: " + fileName);
	if (DEBUG)
	    System.out.println("Getting " + CARDS + " cards " + fileName);
	img = getImage(getDocumentBase(), fileName);
	tracker.addImage(img, 0);
	try  {
	    tracker.waitForID(0);
	}  catch  (InterruptedException  ex)  {
	    // return;
	}
	if (tracker.isErrorAny()) {
	    System.out.println("Error getting " + CARDS + " cards " + fileName);
	    System.exit(1);
	}
	showStatus("Got the images, processing...");
	if (DEBUG)
	    System.out.println("Got " + CARDS + " cards " + fileName);

	int imgX = img.getWidth(this);
	int imgY = img.getHeight(this);
	CARDX = imgX / CELLX;
	CARDY = imgY / (CARDS / CELLX);

	for (int i=0; i< CARDS; i++) {
	    cropFilter = new CropImageFilter(CARDX * (i % CELLX),
		CARDY * (i / CELLX), CARDX, CARDY);
	    fiftyTwo[i] = createImage(
		new FilteredImageSource(img.getSource(), cropFilter));
	}

	//
	// img.flush();		// netscape has problems with this
	//
	showStatus("Finished processing");
	if (DEBUG)
	    System.out.println("done images.MakeCard ");
	//
	CARDX = (CARDX * scale) / 100;
	CARDY = (CARDY * scale) / 100;
	//
	resize(preferredSize());
	//
	layout = new PyramidLayout(CARDX, CARDY);
	crib = new Crib(CARDX, CARDY);
	newColor = new MorphColor(DEBUG);
	setup();
	history = new History(crib, deck, DEBUG);
	hcanvas = new HistoryCanvas(history, DEBUG, this,
		new Dimension((CARDX + PADD) * 2, (CARDY + PADD) * 2));
	rules = new Rules(crib, deck, history, DEBUG);
	pcanvas = new PyramidCanvas(CARDX, CARDY, DEBUG, layout, deck,
			nextX, this, fiftyTwo, rules);
	ccanvas = new CribCanvas(DEBUG, CARDX, CARDY, scale,
			crib, this, fiftyTwo, rules, deck, nextX);
	ibutton = new IdiotButton(jumpy, DEBUG, this,
		new Dimension((CARDX + PADD) * 2, (CARDY + PADD) * 2), scale);
	//
	setLayout(gb);
	gc.fill = GridBagConstraints.NONE;
	gc.gridwidth = GridBagConstraints.RELATIVE;
	gc.gridheight = GridBagConstraints.REMAINDER;
	gc.anchor = GridBagConstraints.NORTHWEST;
	gb.setConstraints(pcanvas, gc);
	add(pcanvas);
	//
	gc.gridwidth = GridBagConstraints.REMAINDER;
	gc.gridheight = 2;
	gc.anchor = GridBagConstraints.NORTHEAST;
	gb.setConstraints(ibutton, gc);
	add(ibutton);
	//
	gc.gridwidth = GridBagConstraints.REMAINDER;
	gc.gridheight = GridBagConstraints.RELATIVE;
	gc.anchor = GridBagConstraints.NORTHEAST;
	gb.setConstraints(ccanvas, gc);
	add(ccanvas);
	//
	gc.fill = GridBagConstraints.BOTH;
	gc.weightx = 0.0;
	gc.weighty = 0.0;
	gc.gridwidth = GridBagConstraints.REMAINDER;
	gc.gridheight = GridBagConstraints.REMAINDER;
	gc.anchor = GridBagConstraints.SOUTHEAST;
	gb.setConstraints(hcanvas, gc);
	add(hcanvas);
	//
	resize(preferredSize());
	ibutton.restart();
	//
    }

    public Dimension minimumSize() {
	return new Dimension((CARDX + PADD) * (ROWS + 2), (CARDY + PADD) * ROWS);
    }

    public Dimension preferredSize() {
	return minimumSize();
    }

    private void setup () {
	setBackground(Color.red);
	deck = new Shuffle(CARDS);
	nextX = newColor.colorRandom();
    }

    synchronized void restart() {
	if (DEBUG)
	    System.out.println("============================== restarting ==============================");
	setup();
	crib.restart();
	pcanvas.restart(deck, nextX);
	ccanvas.restart(deck, nextX);
	ibutton.restart();
	history.restart(deck);
	hcanvas.restart();
	rules.restart(deck);
	repaint();
    }

    public void update(Graphics g) {
	paint(g);
    }

    public void update(boolean pc, boolean cc, boolean hc) {
	if (pc) pcanvas.repaint();
	if (cc) ccanvas.repaint();
	if (hc) hcanvas.update();
    }

    public void paint(Graphics g) {
	if (pcanvas != null)
	    pcanvas.repaint();
	if (ccanvas != null)
	    ccanvas.repaint();
	if (hcanvas != null)
	    hcanvas.repaint();
	if (ibutton != null)
	    ibutton.repaint();
    }
}
